// ympack.c
// YM packer used for the Shymmer music disk
// Mic, 2013

#include <stdio.h>
#include <stdbool.h>
#include <stdint.h>
#include <malloc.h>

uint32_t numFrames, currFrame, loopFrame, dataOffs;
uint32_t numDigiDrums;
uint8_t *ymData, *ymRegStream;

int regs[16];

int main(int argc, char**argv)
{
	FILE *fp;
	size_t outSize = 0;

	if (argc < 3) {
		printf("Usage: ympack input outout\n");
		return 1;
	}

	fp = fopen(argv[1], "rb");
	if (fp == NULL) {
		printf("Failed to open %s\n", argv[1]);
		return 1;
	}

	for (int i = 0; i < 16; i++)
		regs[i] = -1;

	fseek(fp, 0, SEEK_END);
    size_t fileSize = ftell(fp);
    fseek(fp, 0, SEEK_SET);

    ymData = malloc(fileSize);

    fread(ymData, 1, fileSize, fp);
    fclose(fp);

	numDigiDrums = ((uint16_t)ymData[0x14]) << 8;
	numDigiDrums |= ymData[0x15];

	ymRegStream = (uint8_t*)ymData + 0x22;

	while (numDigiDrums) {
		uint32_t sampleBytes = ((uint32_t)ymRegStream[0]) << 24;
		sampleBytes |= ((uint32_t)ymRegStream[1]) << 16;
		sampleBytes |= ((uint32_t)ymRegStream[2]) << 8;
		sampleBytes |= ((uint32_t)ymRegStream[3]);

		ymRegStream += 4;

		ymRegStream += sampleBytes;
		numDigiDrums--;
	}

	while (*ymRegStream++);		// Skip song name
	while (*ymRegStream++);		// Skip author name
	while (*ymRegStream++);		// Skip song comment

	currFrame = 0;

	numFrames = (uint32_t)ymData[0x0E] << 8;
	numFrames += ymData[0x0F];

	loopFrame = (uint32_t)ymData[0x1E] << 8;
	loopFrame += ymData[0x1F];

	fp = fopen(argv[2], "wb");
	if (fp == NULL) {
		printf("Failed to open %s\n", argv[2]);
		return 1;
	}

	for (uint8_t *p = ymData; p < ymRegStream; p++) {
		fputc(*p, fp);
		outSize++;
	}

	printf("ympack: packing %s..", argv[1]);

	size_t outHeaderSize = outSize;
	bool loopHandled = false;

	while (currFrame < numFrames) {
		uint16_t newValues = 0;

		dataOffs = currFrame;
		for (int i = 0; i < 16; i++) {
			if (ymRegStream[dataOffs] != regs[i])
				newValues |= (1 << i);
			dataOffs += numFrames;
		}

		if (currFrame == loopFrame && !loopHandled) {
			loopFrame = outSize - outHeaderSize;
			loopHandled = true;
			newValues = 0xFFFF;
		}

		fputc(newValues & 0xFF, fp);
		fputc(newValues >> 8, fp);
		outSize += 2;

		dataOffs = currFrame;
		for (int i = 0; i < 16; i++) {
			if (newValues & (1 << i)) {
				fputc(ymRegStream[dataOffs], fp);
				outSize++;
			}
			regs[i] = ymRegStream[dataOffs];
			dataOffs += numFrames;
		}

		currFrame++;
	}

	fseek(fp, 0x1C, SEEK_SET);
	fputc(loopFrame >> 8, fp);
	fputc(loopFrame & 0xFF, fp);

	fclose(fp);

	printf(" Done. %d bytes -> %d bytes\n", fileSize, outSize);

	free(ymData);
	return 0;
}


