# SEGA 32X support code for the 68000
# by Chilly Willy

# Shymmer UI stuff by Mic

        .text

# 0x880800 - entry point for reset/cold-start

.include "genvdp.inc"

        .global _start
_start:

# Clear Work RAM
        moveq   #0,d0
        move.w  #0x3FFF,d1
        suba.l  a1,a1
1:
        move.l  d0,-(a1)
        dbra    d1,1b

# Copy initialized variables from ROM to Work RAM
        lea     __text_end,a0
        move.w  #__data_size,d0
        lsr.w   #1,d0
        subq.w  #1,d0
2:
        move.w  (a0)+,(a1)+
        dbra    d0,2b

        lea     __stack,sp              /* set stack pointer to top of Work RAM */

        bsr     init_hardware           /* initialize the console hardware */

        jsr     main                    /* call program main() */
3:
        stop    #0x2700
        bra.b   3b

        .align  64

# 0x880840 - 68000 General exception handler

        move.l  d0,-(sp)
        move.l  4(sp),d0            /* jump table return address */
        sub.w   #0x206,d0           /* 0 = BusError, 6 = AddrError, etc */

        /* handle exception here */

        move.l  (sp)+,d0
        addq.l  #4,sp               /* pop jump table return address */
        rte

        .align  64

# 0x880880 - 68000 Level 4 interrupt handler - HBlank IRQ
        rte

        .align  64

# 0x8808C0 - 68000 Level 6 interrupt handler - VBlank IRQ

        move.l  d0,-(sp)

	move.w	#0,vscroll_line

        move.l  a0,-(sp)
        move.l #vtimer,a0
        addq.l #1,(a0)
        move.l	(sp)+,a0
        
        move.l  vblank,d0
        beq.b   1f
        move.l  a0,-(sp)
        movea.l d0,a0
        jmp     (a0)
1:
        move.l  (sp)+,d0
        rte


# Initialize the MD side to a known state for the game

init_hardware:
        lea     0xC00004,a0
        move.w  #0x8104,(a0)            /* display off, vblank disabled */
        move.w  (a0),d0                 /* read VDP Status reg */

# init joyports
        move.b  #0x40,0xA10009
        move.b  #0x40,0xA1000B
        move.b  #0x40,0xA10003
        move.b  #0x40,0xA10005

        move.w  #0,0xA15128         /* controller 1 */
        move.w  #0,0xA1512A         /* controller 2 */

        move.l  #0,0xA1512C         /* clear the vblank count */

# allow the 68k to access the FM chip
        move.w  #0x0100,0xA11100    /* Z80 assert bus request */
        move.w  #0x0100,0xA11200    /* Z80 deassert reset */

# wait on Mars side
        move.b  #0,0xA15107             /* clear RV - allow SH2 to access ROM */
0:
        cmp.l   #0x4D5F4F4B,0xA15120    /* M_OK */
        bne.b   0b                      /* wait for master ok */
1:
        cmp.l   #0x535F4F4B,0xA15124    /* S_OK */
        bne.b   1b                      /* wait for slave ok */

        move.l  #vert_blank,vblank  /* set vertical blank interrupt handler */
        move.w  #0x2000,sr          /* enable interrupts */
        rts


# void write_byte(void *dst, unsigned char val)
        .global write_byte
write_byte:
        movea.l 4(sp),a0
        move.l  8(sp),d0
        move.b  d0,(a0)
        rts

# void write_word(void *dst, unsigned short val)
        .global write_word
write_word:
        movea.l 4(sp),a0
        move.l  8(sp),d0
        move.w  d0,(a0)
        rts

# void write_long(void *dst, unsigned int val)
        .global write_long
write_long:
        movea.l 4(sp),a0
        move.l  8(sp),d0
        move.l  d0,(a0)
        rts

# unsigned char read_byte(void *src)
        .global read_byte
read_byte:
        movea.l 4(sp),a0
        move.b  (a0),d0
        rts

# unsigned short read_word(void *src)
        .global read_word
read_word:
        movea.l 4(sp),a0
        move.w  (a0),d0
        rts

# unsigned int read_long(void *src)
        .global read_long
read_long:
        movea.l 4(sp),a0
        move.l  (a0),d0
        rts

init_gfx:
	move.l 		#GFXCNTL,a3
	write_vdp_reg 	0,VDP0_PLTT_FULL
	write_vdp_reg 	1,(VDP1_E_VBI + VDP1_E_DISPLAY + VDP1_E_DMA + VDP1_RESERVED)
	write_vdp_reg 	2,(0xc000 >> 10)	/* Screen map a adress */
	write_vdp_reg 	3,(0xc000 >> 10)	/* Window address */
	write_vdp_reg 	4,(0xe000 >> 13)	/* Screen map b address */
	write_vdp_reg 	5,(0xfc00 >>  9)	/* Sprite address */
	write_vdp_reg 	6,0	
	write_vdp_reg	7,0			/* Border color */
	write_vdp_reg	8,1			/* Unused (?) */
	write_vdp_reg	9,1			/* Unused (?) */
	write_vdp_reg	10,1			/* Lines per hblank interrupt */
	write_vdp_reg	11,0			/* Fullscreen vertical scrolling */
	write_vdp_reg	12,(VDP12_SCREEN_V224 + VDP12_SCREEN_H320 + VDP12_PROGRESSIVE)
	write_vdp_reg	13,(0x6000 >> 10)	/* Horizontal scroll address */
	write_vdp_reg	15,2
	write_vdp_reg	16,(VDP16_MAP_V32 + VDP16_MAP_H64)
	write_vdp_reg	17,0
	write_vdp_reg	18,0xff
	rts
	

reset_vscroll_table:
	move.l		#vscroll_table,a0
	move.l		#256,d0
_reset_vscroll_loop:
	move.w		#0,(a0)+
	dbra		d0,_reset_vscroll_loop
	rts

setup_vscroll:
	move.l 		#GFXCNTL,a3
	move.l 		#GFXDATA,a4
	VSCROLL_ADDR 	d0,0x0000
	move.l		d0,(a3)
	move.l		#40-1,d1
_write_vscroll_data:
	move.w		#0,(a4)		
	dbra 		d1,_write_vscroll_data
	rts

update_vscroll_table:
	move.l		#vscroll_table,a0
	move.l		nfo_text_scroll,d1
	and.w		#0x1F,d1
	move.l		#120,d0
0:
	move.w		d1,(a0)+
	dbra		d0,0b
	rts
	
	
        .data

# Put remaining code in data section to lower bus contention for the rom.

        .global do_main
do_main:
        move.b  #1,0xA15107         /* set RV */
        move.b  #2,0xA130F1         /* SRAM disabled, write protected */
        move.b  #0,0xA15107         /* clear RV */

        move.w  0xA15100,d0
        or.w    #0x8000,d0
        move.w  d0,0xA15100         /* set FM - allow SH2 access to MARS hw */
        move.l  #0,0xA15120         /* let Master SH2 run */

	# Decode pattern data
  	move.l 	#FONT_DATA_END,d0
  	sub.l  	#FONT_DATA,d0
  	move.l 	#FONT_DATA,a2
  	move.l 	#0x1000,a3
  	jsr 	lzss_decode_vram

	# Decode pattern data
  	move.l 	#ABOUTBOX_DATA_END,d0
  	sub.l  	#ABOUTBOX_DATA,d0
  	move.l 	#ABOUTBOX_DATA,a2
  	move.l 	#0x0000,a3
  	jsr 	lzss_decode_vram

	# Load palette 
	move.l	#0x0000,a3
	move.l 	#ABOUTBOX_PALETTE,a4
	move.l	#6,d4
	jsr	load_colors

	# Clear map A 
	move.l	#0xE000,a0
	move.l	#160,d0		/* 160 is the space character */
	jsr	clear_map
	
	# Clear map B 
	move.l	#0xC000,a0
	move.l	#160,d0		/* 160 is the space character */
	jsr	clear_map

	jsr	init_gfx
	


main_loop:
        move.w  0xA15120,d0         /* get COMM0 */
        bne.b   handle_req

        nop
        bra.b   main_loop

# process request from Master SH2
handle_req:
        cmpi.w  #0x01FF,d0
        bls     read_sram
        cmpi.w  #0x02FF,d0
        bls     write_sram
        cmpi.w  #0x03FF,d0
        bls     start_music
        cmpi.w  #0x04FF,d0
        bls     toggle_nfo
# unknown command
        move.w  #0,0xA15120         /* done */
        bra.b   main_loop

read_sram:
        move.w  #0x2700,sr          /* disable ints */
        moveq   #0,d1
        moveq   #0,d0
        move.w  0xA15122,d0         /* COMM2 holds offset */
        lea     0x200000,a0
        move.b  #1,0xA15107         /* set RV */
        move.b  #3,0xA130F1         /* SRAM enabled, write protected */
        move.b  1(a0,d0.l),d1       /* read SRAM */
        move.b  #2,0xA130F1         /* SRAM disabled, write protected */
        move.b  #0,0xA15107         /* clear RV */
        move.w  d1,0xA15122         /* COMM2 holds return byte */
        move.w  #0,0xA15120         /* done */
        move.w  #0x2000,sr          /* enable ints */
        bra     main_loop

write_sram:
        move.w  #0x2700,sr          /* disable ints */
        moveq   #0,d1
        move.w  0xA15122,d1         /* COMM2 holds offset */
        lea     0x200000,a0
        move.b  #1,0xA15107         /* set RV */
        move.b  #1,0xA130F1         /* SRAM enabled, write enabled */
        move.b  d0,1(a0,d1.l)       /* write SRAM */
        move.b  #2,0xA130F1         /* SRAM disabled, write protected */
        move.b  #0,0xA15107         /* clear RV */
        move.w  #0,0xA15120         /* done */
        move.w  #0x2000,sr          /* enable ints */
        bra     main_loop

set_rom_bank:
        move.l  a0,d3
        swap    d3
        lsr.w   #4,d3
        andi.w  #3,d3
        move.w  d3,0xA15104         /* set ROM bank select */
        move.l  a0,d3
        andi.l  #0x0FFFFF,d3
        ori.l   #0x900000,d3
        movea.l d3,a1
        rts

start_music:
        bra     main_loop

toggle_nfo:
	move.w  #0x2700,sr
	move.w  #0,0xA15120
	eor.b	#1,display_nfo
	btst.b	#0,display_nfo
	bne	1f
	# Clear map A 
	move.l	#0xE000,a0
	move.l	#160,d0		/* 160 is the space character */
	jsr	clear_map
	# Clear map B 
	move.l	#0xC000,a0
	move.l	#160,d0		/* 160 is the space character */
	jsr	clear_map
	move.w  #0x2000,sr
	bra main_loop
1:
	move.l	#0,nfo_text_scroll
	jsr	show_nfo_box
	jsr	puts_nfo_text
	move.w  #0x2000,sr
	bra	main_loop
	

vert_blank:
        move.l  d1,-(sp)
        move.l  d2,-(sp)

        /* read controllers */
        move.w  0xA15128,d0
        andi.w  #0xF000,d0
        cmpi.w  #0xF000,d0
        beq.b   0f                  /* no pad in port 1 (or mouse) */
        lea     0xA10003,a0
        bsr   	get_pad
        move.w  d2,0xA15128         /* controller 1 current value */
        move.w	pad1_prev_buttons,d1
        eor.w	d2,d1		/* d1 = new buttons */
        and.w	d2,d1
        move.w	d2,pad1_prev_buttons
        btst.b	#0,display_nfo
        beq.b	not_up
        btst.b	#1,d1
        beq.b	not_down
	move.l	nfo_text_scroll,d0
	cmp.l	#82,d0
	bhs.b	not_down
        add.l	#1,nfo_text_scroll
        jsr	puts_nfo_text
        bra	not_up
not_down:
	btst.b	#0,d1
	beq.b	not_up
	move.l	nfo_text_scroll,d0
	cmp.l	#0,d0
	beq.b	not_up
	sub.l	#1,nfo_text_scroll
	jsr	puts_nfo_text
not_up:
0:
        move.w  0xA1512A,d0
        andi.w  #0xF000,d0
        cmpi.w  #0xF000,d0
        beq.b   1f                  /* no pad in port 2 (or mouse) */
        lea     0xA10005,a0
        bsr.b   get_pad
        move.w  d2,0xA1512A         /* controller 2 current value */
1:
        move.l  0xA1512C,d0
        addq.l  #1,d0
        move.l  d0,0xA1512C         /* increment the vblank count */

        tst.w   gen_lvl2
        beq.b   2f
        lea     0xA12000,a0
        move.w  (a0),d0
        ori.w   #0x0100,d0
        move.w  d0,(a0)
2:
        move.l  (sp)+,d2
        move.l  (sp)+,d1
        movea.l (sp)+,a0
        move.l  (sp)+,d0
        rte

# get current pad value
# entry: a0 = pad control port
# exit:  d2 = pad value (0 0 0 1 M X Y Z S A C B R L D U) or (0 0 0 0 0 0 0 0 S A C B R L D U)
get_pad:
        bsr.b   get_input       /* - 0 s a 0 0 d u - 1 c b r l d u */
        move.w  d0,d1
        andi.w  #0x0C00,d0
        bne.b   no_pad
        bsr.b   get_input       /* - 0 s a 0 0 d u - 1 c b r l d u */
        bsr.b   get_input       /* - 0 s a 0 0 0 0 - 1 c b m x y z */
        move.w  d0,d2
        bsr.b   get_input       /* - 0 s a 1 1 1 1 - 1 c b r l d u */
        andi.w  #0x0F00,d0      /* 0 0 0 0 1 1 1 1 0 0 0 0 0 0 0 0 */
        cmpi.w  #0x0F00,d0
        beq.b   common          /* six button pad */
        move.w  #0x010F,d2      /* three button pad */
common:
        lsl.b   #4,d2           /* - 0 s a 0 0 0 0 m x y z 0 0 0 0 */
        lsl.w   #4,d2           /* 0 0 0 0 m x y z 0 0 0 0 0 0 0 0 */
        andi.w  #0x303F,d1      /* 0 0 s a 0 0 0 0 0 0 c b r l d u */
        move.b  d1,d2           /* 0 0 0 0 m x y z 0 0 c b r l d u */
        lsr.w   #6,d1           /* 0 0 0 0 0 0 0 0 s a 0 0 0 0 0 0 */
        or.w    d1,d2           /* 0 0 0 0 m x y z s a c b r l d u */
        eori.w  #0x1FFF,d2      /* 0 0 0 1 M X Y Z S A C B R L D U */
        rts

no_pad:
        move.w  #0xF000,d2
        rts

# read single phase from controller
get_input:
        move.b  #0x00,(a0)
        nop
        nop
        move.b  (a0),d0
        move.b  #0x40,(a0)
        lsl.w   #8,d0
        move.b  (a0),d0
        rts


#################################################
#                                               #
#        Load color data from ROM               #
#                                               #
# Parameters:                                   #
#  a3: CRAM base                                # 
#  a4: color list address                       #
#  d4: number of colors to load                 #
#                                               #
#################################################

load_colors:
	move.l 		#GFXCNTL,a2
	CRAM_ADDR_var 	d0,a3
	move.l 		d0,(a2)

	move.l 		#GFXDATA,a3
	subq.l		#1,d4
_copy_color_data:
	move.w		(a4)+,(a3)
	dbra		d4,_copy_color_data
	rts

#################################################
#                                               #
#        Clear one of the screen maps           #
#                                               #
# Parameters:                                   #
#  a0: Map address                              # 
#  d0: Data to write to each map entry          #
#                                               #
#################################################

clear_map:
	move.l 		#GFXCNTL,a4
	VRAM_ADDR_var	d1,a0
	move.l 		d1,(a4)
	move.l 		#GFXDATA,a3
	move.w		#2047,d1	/* Loop counter (64*32-1) */
_clear_map_loop:
	move.w		d0,(a3)
	dbra		d1,_clear_map_loop
	rts
	
	
#################################################
#                                               #
#       Wait for next VBlank interrupt          #
#                                               #
#################################################

wait_vsync:
	movea.l		#vtimer,a0
	move.l		(a0),a1
_wait_change:
	cmp.l		(a0),a1
	beq		_wait_change
	rts

nfo_text:
.ascii "            Shymmer             "
.ascii "   A YM musicdisk for the 32X   "
.ascii " ------------------------------ "
.ascii "                                "
.ascii "Controls:                       "
.ascii "---------                       "
.ascii "[\x1C/\x1D]:   Navigate the song list/"
.ascii "               Scroll info text "
.ascii "[B]:     Play song              "
.ascii "[Start]: Show/hide info         "
.ascii "                                "
.ascii "Background:                     "
.ascii "-----------                     "
.ascii "I had originally written a YM   "
.ascii "player in TLCS-900/H assembly   "
.ascii "language for the Neo-Geo Pocket "
.ascii "Color, which I later decided to "
.ascii "rewrite in C, while adding some "
.ascii "features that hadn't been possi-"
.ascii "ble on the NGPC due to CPU cycle"
.ascii "contraints. I decided to use the"
.ascii "32X as the target platform for  "
.ascii "this updated version.           "
.ascii "                                "
.ascii "Technical information:          "
.ascii "----------------------          "
.ascii "The 32X was a largely unsuccess-"
.ascii "ful add-on for the Sega Genesis/"
.ascii "Megadrive. It had two Hitachi   "
.ascii "SuperH-2 CPUs in a master-slave "
.ascii "configuration, both clocked at  "
.ascii "~23 MHz.                        "
.ascii "It also contained 256 kB of work"
.ascii "RAM; a 128 kB framebuffer (for  "
.ascii "bitmapped graphics); and a PWM  "
.ascii "audio circuit. The audio circuit"
.ascii "can be set up for either mono or"
.ascii "stereo sample data, and for a   "
.ascii "variety of sampling rates, alth-"
.ascii "ough the sample range decreases "
.ascii "as the sampling rate increases. "
.ascii "                                "
.ascii "For this music disk I handle the"
.ascii "YM2149 (or AY-3-8910) emulation "
.ascii "on the slave SH-2, while the ma-"
.ascii "ster SH-2 and the Megadrive's   "
.ascii "M68k/VDP draws the UI and respo-"
.ascii "nds to gamepad input.           "
.ascii "The YM emulator generates 32 kHz"
.ascii "mono data, which is sent through"
.ascii "a cheap low-pass filter to hide "
.ascii "some aliasing, and finally writ-"
.ascii "ten to the audio circuit using a"
.ascii "DMA channel.                    "
.ascii "                                "
.ascii "The YM playback code supports   "
.ascii "YM5 and YM6 (everything else is "
.ascii "handled as YM5), including YM   "
.ascii "effects (SID-Voice and DigiDrums"
.ascii "are working, Sync-Buzz is imple-"
.ascii "mented but untested).           "
.ascii "                                "
.ascii "The YM files would take up appr-"
.ascii "oximately 5 MB when unpacked,   "
.ascii "but the 32X can only access 4 MB"
.ascii "of ROM. So I implemented my own "
.ascii "packing scheme, since I wanted  "
.ascii "to avoid doing LHA-unpacking du-"
.ascii "ring playback, and the amount of"
.ascii "RAM available on the 32X makes  "
.ascii "it impossible to fully unpack   "
.ascii "some of the files into RAM ahead"
.ascii "of playback.                    " 
.ascii "                                "
.ascii "User interface:                 "
.ascii "---------------                 "
.ascii "The UI was designed to resemble "
.ascii "a TOS/GEM environment, although "
.ascii "I took some liberities due to   "
.ascii "the limited resolution I had to "
.ascii "work with and the information I "
.ascii "wanted to fit on screen.        "
.ascii "All of the UI is drawn using the"
.ascii "master SH-2 on the 32X, except  "
.ascii "for this info text box which is "
.ascii "handled by the Megadrive's M68k "
.ascii "and VDP.                        "
.ascii "The 3x5 grid at the bottom-right"
.ascii "of the screen indicates for each"
.ascii "channel whether it's:           " 
.ascii " ..generating tone output.      "
.ascii " ..generating noise output.     "
.ascii " ..using the envelope generator."
.ascii " ..using effect #1.             "
.ascii " ..using effect #2.             "
.ascii "                                "
.ascii "/Mic                            "
dc.b  0

.align 2

puts_nfo_text:
	move.l		a2,-(sp)
	move.l		a3,-(sp)
	move.l		d3,-(sp)
	move.l		d4,-(sp)
	move.l		d5,-(sp)
	
	move.l 		#GFXCNTL,a2
	move.l 		#GFXDATA,a3
	move.l		nfo_text_scroll,d0
	lsl.l		#5,d0
	add.l		#nfo_text,d0
	move.l		d0,a0
	moveq.l		#15-1,d4	/* rows of text to show */
	move.l		#0xC308,d3
_nfo_rows:
	moveq.l		#32-1,d5	/* columns of text to show */
	VRAM_ADDR_var 	d0, d3
	move.l		d0,(a2)		/* set vram address */
	moveq.l		#0,d0
_nfo_columns:
	move.b		(a0)+,d0	/* read character */
	add.b		#128,d0		/* font starts at character 128 */
	move.w		d0,(a3)
	dbra		d5,_nfo_columns
	add.l		#128,d3
	dbra		d4,_nfo_rows

	add.l		#32,d3
	VRAM_ADDR_var 	d0, d3
	move.l		d0,(a2)
	move.w		#128+29,(a3)	/* down arrow */
	
	move.l		(sp)+,d5
	move.l		(sp)+,d4
	move.l		(sp)+,d3
	move.l		(sp)+,a3
	move.l		(sp)+,a2
	rts


show_nfo_box:
	move.l 		#GFXCNTL,a2
	move.l 		#GFXDATA,a3
	move.l		#0xE000,d3	/* map vram address */
	move.l		#ABOUTBOX_NAMETABLE,a0
	move.l		#24-1,d4	/* row counter */
_show_nfo_box_rows:
	VRAM_ADDR_var 	d0, d3
	move.l		d0,(a2)		/* set vram address */
	move.l		#40-1,d5
_show_nfo_box_columns:
	move.w 		(a0)+,(a3)
	dbra		d5,_show_nfo_box_columns
	add.l		#128,d3
	dbra		d4,_show_nfo_box_rows
	rts
	

# Global variables for 68000

        .align  4

vblank:
        dc.l    0

        .global gen_lvl2
gen_lvl2:
        dc.w    0

        .global cd_ok
cd_ok:
        dc.w    0

vtimer:	.long 0
nfo_text_scroll: .long 0
display_nfo: .long 0

vscroll_line: .long 0
vscroll_table: .space 256*2
pad1_buttons: dc.w 0
pad1_prev_buttons: dc.w 0

        .align  4
